import tatsu
from ltn_imp.parsing.expressions import *


class FOLSemantics:
    def start(self, ast):
        return ast
    
    def expression(self, ast):
        return ast
    
    def equivalence(self, ast):
        return IffExpression(ast[0], ast[2])
    
    def implication(self, ast):
        return ImpExpression(ast[0], ast[2])
    
    def disjunction(self, ast):
        return OrExpression(ast[0], ast[2])
    
    def conjunction(self, ast):
        return AndExpression(ast[0], ast[2])
    
    def negation(self, ast):
        return NegatedExpression(ast[1])
    
    def quantified(self, ast):
        quantifier = ExistsExpression if ast[0] == 'exists' else ForallExpression
        return quantifier(ast[1], ast[3])
    
    def predicate(self, ast):
        name = ast[0]  # The name of the predicate
        variables = ast[1]  # The variable within the predicate
        if type(variables) != list:
            variables = [variables]
        return ApplicationExpression(name, variables)
    
    def variable(self, ast):
        return VariableExpression(variable=ast)
    
    def constant(self, ast):
        return ConstantExpression(constant=ast)
    
    def addition(self, ast):
        return AdditionExpression(ast[0], ast[2])
    
    def subtraction(self, ast):
        return SubtractionExpression(ast[0], ast[2])
    
    def multiplication(self, ast):
        return MultiplicationExpression(ast[0], ast[2])
    
    def division(self, ast):
        return DivisionExpression(ast[0], ast[2])
    
    def less_than(self, ast):
        return LessThanExpression(ast[0], ast[2])
    
    def more_than(self, ast):
        return MoreThanExpression(ast[0], ast[2])
    
    def less_equal(self, ast):
        return LessEqualExpression(ast[0], ast[2])
    
    def more_equal(self, ast):
        return MoreEqualExpression(ast[0], ast[2])

    def equality(self, ast):
        return EqualityExpression(ast[0], ast[2])
    
    def direct(self, ast):
        return DirectEqualityExpression(ast[0], ast[2])
    
    def negative(self, ast):
        return NegativeExpression(ast[1])
    
    def index(self, ast):
        return IndexExpression(ast[0], ast[1])

class LTNParser():
    def __init__(self, path):
        with open(path, 'r') as file:
            grammar = file.read()

        self.parser = tatsu.compile(grammar)

    def parse(self, text):
        return self.parser.parse(text, semantics=FOLSemantics())

    def __call__(self, text):
        return self.parse(text)


def collect_variables(expr):
    variables = set()

    if isinstance(expr, VariableExpression):
        variables.add(expr)
    elif isinstance(expr, BinaryExpression):
        variables.update(collect_variables(expr.left))
        variables.update(collect_variables(expr.right))
    elif isinstance(expr, NegatedExpression):
        variables.update(collect_variables(expr.term))
    elif isinstance(expr, NegativeExpression):
        variables.update(collect_variables(expr.term))
    elif isinstance(expr, QuantifiedExpression):
        for variable in expr.variables_:
            variables.add(variable)
        variables.update(collect_variables(expr.term))
    elif isinstance(expr, ApplicationExpression):
        for arg in expr.args:
            variables.update(collect_variables(arg))
    elif isinstance(expr, IndexExpression):
        variables.update(collect_variables(expr.variable))
        variables.update(collect_variables(expr.feature))

    return variables

class Expression:
    def variables(self):
        return collect_variables(self)
    
    
class BinaryExpression(Expression):
    def __init__(self, left, operator, right):
        self.left = left
        self.operator = operator
        self.right = right

    def __repr__(self):
        return f"<{self.__class__.__name__}>: ({self.left} {self.operator} {self.right})"
    
    def __str__(self):
        return f"({self.left} {self.operator} {self.right})"

class AndExpression(BinaryExpression):
    def __init__(self, left, right):
        super().__init__(left, '&', right)

class OrExpression(BinaryExpression):
    def __init__(self, left, right):
        super().__init__(left, '|', right)

class ImpExpression(BinaryExpression):
    def __init__(self, left, right):
        super().__init__(left, '->', right)

class IffExpression(BinaryExpression):
    def __init__(self, left, right):
        super().__init__(left, '<->', right)

class NegatedExpression(Expression):
    def __init__(self, term):
        self.term = term

    def __repr__(self):
        return f"<{self.__class__.__name__}>: ~({self.term})"
    
    def __str__(self):
        return f"~({self.term})"
    
class NegativeExpression(Expression):
    def __init__(self, term):
        self.term = term

    def __repr__(self):
        return f"<{self.__class__.__name__}>: -({self.term})"
    
    def __str__(self):
        return f"-({self.term})"

class EqualityExpression(BinaryExpression):
    def __init__(self, left, right):
        super().__init__(left, '=', right)

class  DirectEqualityExpression(BinaryExpression):
    def __init__(self, left, right):
        super().__init__(left, '==', right)

class AdditionExpression(BinaryExpression):
    def __init__(self, left, right):
        super().__init__(left, '+', right)

class SubtractionExpression(BinaryExpression):
    def __init__(self, left, right):
        super().__init__(left, '-', right)

class MultiplicationExpression(BinaryExpression):
    def __init__(self, left, right):
        super().__init__(left, '*', right)

class DivisionExpression(BinaryExpression):
    def __init__(self, left, right):
        super().__init__(left, '/', right)

class LessThanExpression(BinaryExpression):
    def __init__(self, left, right):
        super().__init__(left, '<', right)

class MoreThanExpression(BinaryExpression):
    def __init__(self, left, right):
        super().__init__(left, '>', right)

class LessEqualExpression(BinaryExpression):
    def __init__(self, left, right):
        super().__init__(left, '<=', right)

class MoreEqualExpression(BinaryExpression):
    def __init__(self, left, right):
        super().__init__(left, '>=', right)

class QuantifiedExpression(Expression):
    def __init__(self, quantifier, variables, term):
        self.quantifier = quantifier
        self.variables_ = variables if isinstance(variables, list) else [variables]
        self.term = term

    def __repr__(self):
        temp = [ str(variable) for variable in self.variables_]
        return f"<{self.__class__.__name__}>: {self.quantifier} {temp}. ({self.term})"
    
    def __str__(self):
        temp = [ str(variable) for variable in self.variables_]
        return f"{self.quantifier} {temp}. ({self.term})"

class ExistsExpression(QuantifiedExpression):
    def __init__(self, variable, term):
        super().__init__('∃', variable, term)

class ForallExpression(QuantifiedExpression):
    def __init__(self, variable, term):
        super().__init__('∀', variable, term)

class ApplicationExpression(Expression):
    def __init__(self, function, args):
        self.function = function  # The function or predicate being applied
        self.args = args

    def __repr__(self):
        args = ", ".join(map(str, self.args))
        return f"<{self.__class__.__name__}>: {self.function}({args})"
    
    def __str__(self):
        args = ", ".join(map(str, self.args))
        return f"{self.function}({args})"
class VariableExpression(Expression):
    def __init__(self, variable):
        self.variable = variable

    def __repr__(self):
        return f"<{self.__class__.__name__}>: {self.variable}"
    
    def __str__(self):
        return self.variable
    
    def __eq__(self, value: object) -> bool:
        if isinstance(value, VariableExpression):
            return self.variable == value.variable
        return False

    def __hash__(self) -> int:
        return hash(self.variable)

class ConstantExpression(Expression):
    def __init__(self, constant):
        self.constant = constant

    def __repr__(self):
        return self.constant
    
class IndexExpression(Expression):
    def __init__(self, variable, feature):
        self.variable = variable
        self.feature = feature

    def __repr__(self):
        return f"<{self.__class__.__name__}>: {self.variable}[{self.feature}]"
    
    def __str__(self):
        return f"{self.variable}[{self.feature}]"


from abc import ABC, abstractmethod
import torch

class AggregationOperator(ABC):
    @abstractmethod
    def __call__(self, xs, dim=None, keepdim=False, mask=None):
        pass

class AggregMin(AggregationOperator):
    def __call__(self, xs, dim=None, keepdim=False, mask=None):
        if mask is not None:
            xs = torch.where(~mask, torch.tensor(float('inf'), dtype=xs.dtype, device=xs.device), xs)
        result = torch.amin(xs, dim=dim, keepdim=keepdim)
        return result

class AggregPMean(AggregationOperator):
    def __init__(self, p=2):
        self.p = p

    def __call__(self, xs, dim=None, keepdim=False, mask=None):
        if mask is not None:
            # Apply mask to exclude certain values
            xs = xs * mask
            sum_p = torch.sum(xs ** self.p, dim=dim, keepdim=keepdim)
            count_p = torch.sum(mask, dim=dim, keepdim=keepdim)
        else:
            sum_p = torch.sum(xs ** self.p, dim=dim, keepdim=keepdim)
            count_p = xs.size(dim) if dim is not None else xs.numel()
        result = (sum_p / count_p) ** (1 / self.p)
        return result

class AggregPMeanError(AggregationOperator):
    def __init__(self, p=2):
        self.p = p

    def __call__(self, xs, dim=None, keepdim=False, mask=None):
        if mask is not None:
            xs = torch.where(~mask, torch.tensor(0.0, dtype=xs.dtype, device=xs.device), xs)
            sum_p = torch.sum((1 - xs) ** self.p, dim=dim, keepdim=keepdim)
            count_p = torch.sum(mask, dim=dim, keepdim=keepdim)
        else:
            sum_p = torch.sum((1 - xs) ** self.p, dim=dim, keepdim=keepdim)
            count_p = xs.size(dim) if dim is not None else xs.numel()
        result = 1 - (sum_p / count_p) ** (1 / self.p)
        return result
    
class SatAgg:
    def __init__(self, agg_op = AggregPMeanError(p=2)):
        if not isinstance(agg_op, AggregationOperator):
            raise TypeError("agg_op must be an instance of AggregationOperator")
        self.agg_op = agg_op

    def __call__(self, *closed_formulas):
        # Collect the truth values from the closed formulas
        truth_values = [torch.tensor(cf, dtype=torch.float32, device=cf.device) if not isinstance(cf, torch.Tensor) else cf for cf in closed_formulas]
        
        # Stack the truth values into a single tensor
        truth_values = torch.stack(truth_values)
        # Apply the aggregation operator
        return self.agg_op(truth_values, dim=0)



from abc import ABC, abstractmethod
from ltn_imp.fuzzy_operators.aggregators import *

# Define the available aggregation operators
agg_op_map = {
    "min": AggregMin,
    "pmean": AggregPMean,
    "pmean_error": AggregPMeanError
}

# Abstract class for quantifiers
class Quantifier(ABC):
    def __init__(self, method):
        if method not in agg_op_map:
            raise ValueError(f"Unknown aggregation operator: {method}")
        self.agg_op = agg_op_map[method]()  # Initialize the aggregation operator

    @abstractmethod
    def __call__(self, truth_values, dim=None):
        pass

class ForallQuantifier(Quantifier):
    def __init__(self, method="pmean_error"):
        super().__init__(method)

    def __call__(self, truth_values, dim=None):
        if not isinstance(truth_values, torch.Tensor):
            truth_values = torch.tensor(truth_values, dtype=torch.float32)
        return self.agg_op(truth_values, dim=dim)

class ExistsQuantifier(Quantifier):
    def __init__(self, method="pmean"):
        super().__init__(method)

    def __call__(self, truth_values, dim=None):
        if not isinstance(truth_values, torch.Tensor):
            truth_values = torch.tensor(truth_values, dtype=torch.float32)
        return self.agg_op(truth_values, dim=dim)


import torch.nn as nn
import inspect

class ModuleFactory:

    def __init__(self, converter):
        self.converter = converter
    
    def get_functionality(self, expression):
        return self.converter(expression)

    def create_module(self, name, params, functionality):
        functionality = self.get_functionality(functionality)
        
        # Define the forward method dynamically
        def forward(self, *args):
            local_vars = dict(zip(params, args))
            return functionality(local_vars)

        # Create a signature with the given parameters
        new_params = [inspect.Parameter('self', inspect.Parameter.POSITIONAL_OR_KEYWORD)] + \
                     [inspect.Parameter(param, inspect.Parameter.POSITIONAL_OR_KEYWORD) for param in params]
        
        new_sig = inspect.Signature(parameters=new_params)
        
        forward.__signature__ = new_sig

        def __call__(self, *args):
            return self.forward(*args)
        
        # Create a new class with the dynamically defined forward method
        module_class = type(str(name), (nn.Module,), {
            'forward': forward,
            '__call__': __call__
        })
        
        self.converter.predicates[name] = module_class()

        return module_class


import torch
import inspect

class Predicate:
    def __init__(self, model: torch.nn.Module):
        self.model = model

    def forward(self, *args, **kwargs):
        # Get the signature of the model's forward method
        signature = inspect.signature(self.model.forward)
        parameters = signature.parameters
        
        # Extract only the required parameters, ignoring 'self'
        required_params = [p for p in parameters.values() if p.default == inspect.Parameter.empty and p.name != 'self']

        # Check if the number of provided arguments matches the required ones
        if len(args) == len(required_params):
            return self.model(*args, **kwargs)
        else:
            # Match args to the model's parameters
            bound_args = signature.bind_partial(*args, **kwargs)
            bound_args.apply_defaults()
            return self.model(*bound_args.args, **bound_args.kwargs)

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)



import torch
import torch.nn as nn

class LogicTensorNetwork:
    def __init__(self, arguments, model):
        self.args = arguments 
        self.model = model     

    def __call__(self, *args):
        if len(args) != len(self.args):
            raise ValueError(f"Expected {len(self.args)} arguments, but got {len(args)}.")
        
        input_tensor = torch.cat(args, dim=1)
        return self.model(input_tensor)
    
    def parameters(self):
        return self.model.parameters()
    
    def float(self):
        self.model = self.model.float() 
        return self
    
    def to(self, device):
        self.model = self.model.to(device)  # Move model to the specified device
        return self 
    
    def eval(self):
        self.model = self.model.eval()
        return self

    def train(self):
        self.model = self.model.train()
        return self
    
    def forward(self, *args):
        return self(*args)
    
    def state_dict(self):
        return self.model.state_dict()
    
    def load_state_dict(self, state_dict):
        self.model.load_state_dict(state_dict)
        return self
    
    def __repr__(self):
        return f"<LogicTensorNetwork: {self.args}>"

class NNFactory:
    def __init__(self):
        pass
    
    def __call__(self, arguments, layers, activations, regularizations):
        network_layers = []

        for (in_size, out_size), activation, regularization_list in zip(layers, activations, regularizations):
            # Add linear layer
            linear_layer = nn.Linear(in_size, out_size)
            network_layers.append(linear_layer)
            
            # Apply regularizations from the list
            if regularization_list:
                for regularization in regularization_list:
                    reg_layers = self._get_regularization(regularization, linear_layer)
                    if reg_layers is not None:
                        network_layers.extend(reg_layers)
            
            # Add activation function
            if activation is not None:
                network_layers.append(self._get_activation(activation))
        
        model = nn.Sequential(*network_layers)
        return LogicTensorNetwork(arguments, model)
        
    def _get_activation(self, activation):
        activations = {
            'relu': nn.ReLU(),
            'sigmoid': nn.Sigmoid(),
            'tanh': nn.Tanh(),
            'elu': nn.ELU(),
            'leaky_relu': nn.LeakyReLU(),
            'softmax': nn.Softmax(dim=1),
            'identity': nn.Identity()
        }
        return activations.get(activation, nn.Identity())  # Use Identity if activation is None

    def _get_regularization(self, regularization, linear_layer):
        reg_layers = []
        
        if regularization is None:
            return reg_layers
        
        # Add dropout layer if specified
        if 'dropout' in regularization:
            reg_layers.append(nn.Dropout(regularization['dropout']))
        
        # Add batch normalization if specified
        if 'batch_norm' in regularization and regularization['batch_norm']:
            reg_layers.append(nn.BatchNorm1d(linear_layer.out_features))
        
        # Add layer normalization if specified
        if 'layer_norm' in regularization and regularization['layer_norm']:
            reg_layers.append(nn.LayerNorm(linear_layer.out_features))
        
        return reg_layers


from abc import ABC, abstractmethod
import torch

# Base class for all connectives
class Connective(ABC):
    @abstractmethod
    def forward(self, *args):
        pass

    def __call__(self, *args):
        return self.forward(*args)

# Base class for binary connectives
class BinaryConnective(Connective):
    def __init__(self, implementation):
        self.implementation = implementation

    def forward(self, a, b):
        return self.implementation(a, b)

# Base class for unary operations
class UnaryConnective(Connective):
    def __init__(self, implementation):
        self.implementation = implementation

    def forward(self, a):
        return self.implementation(a)

# And Connective and its subclasses
class AndConnective(BinaryConnective):
    def __init__(self, implementation, stable=True):
        self.stable = stable
        super().__init__(implementation)

class MinAndConnective(AndConnective):
    def __init__(self):
        super().__init__(self.implementation)

    def implementation(self, a, b):
        return torch.minimum(a, b)

class ProdAndConnective(AndConnective):
    def __init__(self, stable=True):
        super().__init__(self.implementation, stable=stable)

    def implementation(self, a, b):
        eps = 1e-4
        if self.stable:
            a = (1 - eps) * a + eps
            b = (1 - eps) * b + eps
        return torch.mul(a, b)

class LukAndConnective(AndConnective):
    def __init__(self):
        super().__init__(self.implementation)

    def implementation(self, a, b):
        return torch.maximum(a + b - 1, torch.zeros_like(a))

class DefaultAndConnective(MinAndConnective):
    pass

# Or Connective and its subclasses
class OrConnective(BinaryConnective):
    def __init__(self, implementation, stable=True):
        self.stable = stable
        super().__init__(implementation)

class MaxOrConnective(OrConnective):
    def __init__(self):
        super().__init__(self.implementation)

    def implementation(self, a, b):
        return torch.maximum(a, b)

class ProbSumOrConnective(OrConnective):
    def __init__(self, stable=True):
        super().__init__(self.implementation, stable=stable)

    def implementation(self, a, b):
        eps = 1e-4
        if self.stable:
            a = (1 - eps) * a
            b = (1 - eps) * b
        return a + b - a * b

class LukOrConnective(OrConnective):
    def __init__(self):
        super().__init__(self.implementation)

    def implementation(self, a, b):
        return torch.minimum(a + b, torch.ones_like(a))

class DefaultOrConnective(MaxOrConnective):
    pass

# Implies Connective and its subclasses
class ImpliesConnective(BinaryConnective):
    def __init__(self, implementation, stable=True):
        self.stable = stable
        super().__init__(implementation)

class KleeneDienesImpliesConnective(ImpliesConnective):
    def __init__(self):
        super().__init__(self.implementation)

    def implementation(self, a, b):
        # Ensure a and b have the same shape and are at least 2D tensors
        if a.ndim == 1:
            a = a.unsqueeze(1)
        if b.ndim == 1:
            b = b.unsqueeze(1)

        # Perform the operation
        result = torch.maximum(1. - a, b)
        return result

class GodelImpliesConnective(ImpliesConnective):
    def __init__(self):
        super().__init__(self.implementation)

    def implementation(self, a, b):
        return torch.where(a <= b, torch.ones_like(a), b)

class ReichenbachImpliesConnective(ImpliesConnective):
    def __init__(self, stable=True):
        super().__init__(self.implementation, stable=stable)

    def implementation(self, a, b):
        eps = 1e-4
        if self.stable:
            a = (1 - eps) * a + eps
            b = (1 - eps) * b
        return 1. - a + a * b

class GoguenImpliesConnective(ImpliesConnective):
    def __init__(self, stable=True):
        super().__init__(self.implementation, stable=stable)

    def implementation(self, a, b):
        eps = 1e-4
        if self.stable:
            a = (1 - eps) * a + eps
        return torch.where(a <= b, torch.ones_like(a), b / a)

class LukImpliesConnective(ImpliesConnective):
    def __init__(self):
        super().__init__(self.implementation)

    def implementation(self, a, b):
        return torch.minimum(1.0 - a + b, torch.ones_like(a))

class DefaultImpliesConnective(KleeneDienesImpliesConnective):
    pass

# Iff Connective and its subclasses
class IffConnective(BinaryConnective):
    def __init__(self, implementation):
        super().__init__(implementation)

class DefaultIffConnective(IffConnective):
    def __init__(self):
        super().__init__(self.implementation)

    def implementation(self, a, b):
        return 1 - torch.abs(a - b)



# Not Connective and its subclasses
class NotConnective(UnaryConnective):
    def __init__(self, implementation):
        super().__init__(implementation)

class StandardNotConnective(NotConnective):
    def __init__(self):
        super().__init__(self.implementation)

    def implementation(self, a):
        return 1 - a

class GodelNotConnective(NotConnective):
    def __init__(self):
        super().__init__(self.implementation)

    def implementation(self, a):
        return torch.eq(a, 0.).float()

class DefaultNotConnective(StandardNotConnective):
    pass

class NegativeConnective(UnaryConnective):
    def __init__(self, implementation):
        super().__init__(implementation)

class DefaultNegativeConnective(NegativeConnective):
    def __init__(self):
        super().__init__(self.implementation)

    def implementation(self, a):
        return torch.sub(0,a)

class EqConnective(BinaryConnective):
    def __init__(self, implementation):
        super().__init__(implementation)

class TanEqConnective(EqConnective): # This returns a tensor of the form [ [0], [1], [0], [1], [0], [1], ... ]
    def __init__(self):
        super().__init__(self.implementation)

    def implementation(self, a, b):
        # Ensure a and b have the same shape and are at least 2D tensors
        if a.ndim == 1:
            a = a.unsqueeze(1)
        if b.ndim == 1:
            b = b.unsqueeze(1)

        # Perform the operation
        result = 1 - torch.tanh(2 * torch.abs(torch.sub(a,b)))**2        
        return result
    
class SqrtEqConnective(EqConnective):  # This returns a tensor in the form [ 0, 1, 0, 1, 0, 1, ...]
    def __init__(self):
        super().__init__(self.implementation)

    def implementation(self, a, b):
        
        # Ensure a and b have the samzaze shape and are at least 2D tensors
        if a.ndim == 0:
            a = a.unsqueeze(0)
        if b.ndim == 0:
            b = b.unsqueeze(0)
        if a.ndim == 1:
            a = a.unsqueeze(1)
        if b.ndim == 1:
            b = b.unsqueeze(1)

        alpha = 0.005  # You can adjust the value of alpha as needed
        result = torch.exp(-alpha * torch.sqrt(torch.sum(torch.square(a - b), dim=1)))

        return result
    
DefaultEqConnective = SqrtEqConnective

# LessThan Connective
class LessThanConnective(BinaryConnective):
    def __init__(self, k=10):
        self.k = k
        super().__init__(self.implementation)

    def normalize_min_max(self, tensor1, tensor2):
        
        if tensor2.dim() == 0:
            tensor2 = tensor2.unsqueeze(0)

        combined_min = torch.min(torch.cat((tensor1, tensor2)))
        combined_max = torch.max(torch.cat((tensor1, tensor2)))
        tensor1_normalized = (tensor1 - combined_min) / (combined_max - combined_min)
        tensor2_normalized = (tensor2 - combined_min) / (combined_max - combined_min)
        return tensor1_normalized, tensor2_normalized

    def implementation(self, tensor1, tensor2):
        tensor1_normalized, tensor2_normalized = self.normalize_min_max(tensor1, tensor2)
        result = torch.sigmoid(self.k * (tensor2_normalized - tensor1_normalized))
        return result

    
class DefaultLessThanConnective(LessThanConnective):
    pass

# MoreThan Connective
class MoreThanConnective(BinaryConnective):
    def __init__(self, k=10):
        self.k = k
        super().__init__(self.implementation)

    def normalize_min_max(self, tensor1, tensor2):
        
        if tensor2.dim() == 0:
            tensor2 = tensor2.unsqueeze(0)

        combined_min = torch.min(torch.cat((tensor1, tensor2)))
        combined_max = torch.max(torch.cat((tensor1, tensor2)))
        tensor1_normalized = (tensor1 - combined_min) / (combined_max - combined_min)
        tensor2_normalized = (tensor2 - combined_min) / (combined_max - combined_min)
        return tensor1_normalized, tensor2_normalized

    def implementation(self, tensor1, tensor2):
        tensor1_normalized, tensor2_normalized = self.normalize_min_max(tensor1, tensor2)
        result = torch.sigmoid(self.k * (tensor1_normalized - tensor2_normalized))
        return result


class DefaultMoreThanConnective(MoreThanConnective):
 pass

# Add Connective
class AddConnective(BinaryConnective):
    def __init__(self):
        super().__init__(self.implementation)

    def implementation(self, tensor1, tensor2):
        return torch.add(tensor1, tensor2).float()

class DefaultAddConnective(AddConnective):
 pass

# Subtract Connective
class SubtractConnective(BinaryConnective):
    def __init__(self):
        super().__init__(self.implementation)

    def implementation(self, tensor1, tensor2):
        return torch.sub(tensor1, tensor2).float()

class DefaultSubtractConnective(SubtractConnective):
    pass

# Multiply Connective
class MultiplyConnective(BinaryConnective):
    def __init__(self):
        super().__init__(self.implementation)

    def implementation(self, tensor1, tensor2):
        return torch.mul(tensor1, tensor2).float()

class DefaultMultiplyConnective(MultiplyConnective):
    pass
# Divide Connective
class DivideConnective(BinaryConnective):
    def __init__(self):
        super().__init__(self.implementation)

    def implementation(self, tensor1, tensor2):
        return torch.div(tensor1, tensor2).float()

class DefaultDivideConnective(DivideConnective):
    pass


class LessThanOrEqualConnective(BinaryConnective):
    def __init__(self, k=5):
        self.k = k
        super().__init__(self.implementation)

    def implementation(self, tensor1, tensor2):
        return DefaultOrConnective()(
            DefaultLessThanConnective(self.k)(tensor1, tensor2),
            DefaultEqConnective()(tensor1, tensor2)
        )

class DefaultLessThanOrEqualConnective(LessThanOrEqualConnective):
    pass

# MoreThanOrEqual Connective
class MoreThanOrEqualConnective(BinaryConnective):
    def __init__(self, k=5):
        self.k = k
        super().__init__(self.implementation)

    def implementation(self, tensor1, tensor2):
        return DefaultOrConnective()(
            DefaultLessThanConnective(self.k)(tensor2, tensor1),
            DefaultEqConnective()(tensor1, tensor2)
        )

class DefaultMoreThanOrEqualConnective(MoreThanOrEqualConnective):
    pass


class Visitor:
    def visit(self, node):
        for cls in type(node).mro():
            method_name = 'visit_' + cls.__name__
            if hasattr(self, method_name):
                visitor = getattr(self, method_name)
                return visitor(node)
        raise ValueError(f"No visitor for {node} of type {type(node)} found in {self}.")



def accept(self, visitor: Visitor):
    assert isinstance(visitor, Visitor), f"Expected a {Visitor.__name__} object, got {type(visitor)}"
    return visitor.visit(self)


def make_visitable(cls: type, accept_method_name: str = 'accept'):
    setattr(cls, accept_method_name, accept)



import torch 
from ltn_imp.fuzzy_operators.aggregators import SatAgg
from ltn_imp.automation.data_loaders import CombinedDataLoader, LoaderWrapper
from ltn_imp.parsing.expressions import LessThanExpression, MoreThanExpression, EqualityExpression
from ltn_imp.parsing.parser import LTNConverter
from ltn_imp.parsing.ancillary_modules import ModuleFactory
from ltn_imp.automation.network_factory import NNFactory
import yaml

sat_agg_op = SatAgg()

class KnowledgeBase:
    def __init__(self, yaml_file, device = "cpu"):

        if device == "cuda" and torch.cuda.is_available():
            self.device = torch.device("cuda")
        elif device == "mps" and torch.backends.mps.is_available():
            self.device = torch.device("mps")
        else:
            self.device = torch.device("cpu")
    
        print(f"Using device: {self.device}")
        
        with open(yaml_file, "r") as file:
            config = yaml.safe_load(file)
        self.config = config
        self.factory = NNFactory()
        self.scalers = {}

        self.set_loaders()
        self.set_val_loaders()
        self.set_test_loaders()

        self.constant_mapping = self.set_constant_mapping()
        self.set_predicates()
        self.set_converter()
        self.set_ancillary_rules()
        self.set_rules()
        self.set_rule_weights()
        self.set_rule_to_data_loader_mapping()

    def set_converter(self):
        self.converter = LTNConverter(yaml=self.config, predicates=self.predicates, device=self.device, scalers=self.scalers) 

    def set_constant_mapping(self):
        constants = self.config.get("constants", [])
        constant_mapping = {}
        for const in constants:
            for name, value in const.items():
                if isinstance(value, list):
                    constant_mapping[name] = torch.tensor(value, dtype=torch.float32)
                else:
                    constant_mapping[name] = torch.tensor([value], dtype=torch.float32)

        return constant_mapping

    def evaluate_layer_size(self, layer_size, features_dict, instance_names):
        in_size_str, out_size_str = layer_size
        for instance_name in instance_names:
            feature_count = len(features_dict[instance_name])
            if isinstance(in_size_str, str):
                in_size_str = in_size_str.replace(instance_name, str(feature_count))
            if isinstance(out_size_str, str):
                out_size_str = out_size_str.replace(instance_name, str(feature_count))

        in_size = eval(in_size_str) if isinstance(in_size_str, str) else in_size_str
        out_size = eval(out_size_str) if isinstance(out_size_str, str) else out_size_str
        return in_size, out_size

    def set_predicates(self):
        features = self.config["features"]
        self.predicates = {}

        for predicate_name, predicate_info in self.config["predicates"].items():
            args = predicate_info["args"]
            structure = predicate_info["structure"]
            
            layers = []
            activations = []
            regularizations = []

            for layer in structure['layers']:
                layer_type = list(layer.keys())[0]
                layer_size = layer[layer_type]
                activation = layer.get('activation', None)
                regularization = layer.get('regularization', [])
                
                in_size, out_size = self.evaluate_layer_size(layer_size, features, args)
                layers.append((in_size, out_size))
                activations.append(activation)
                regularizations.append(regularization)

            arguments = predicate_info["args"]

            network = self.factory(
                arguments=arguments,
                layers=layers,
                activations=activations,
                regularizations=regularizations
            )

            self.predicates[predicate_name] = network.float()

        for predicate in self.predicates.values():
            predicate.to(self.device)

    def set_rules(self):
        self.rules = [self.converter(rule["rule"]) for rule in self.config["constraints"]]

    def set_rule_weights(self):
        self.rule_weights = [int(rule.get("weight", 1)) for rule in self.config["constraints"]]

    def set_ancillary_rules(self):
        if "knowledge" not in self.config:
            return
        for anchor in self.config["knowledge"]:
            name = anchor["rule"]
            params = anchor["args"]
            functionality = anchor["clause"]
            ModuleFactory(self.converter).create_module(name, params, functionality)

    def set_loaders(self):
        self.loaders = []
        features = self.config["features"]
        for dataset in self.config["train"]:
            dataset = self.config["train"][dataset]
            type = dataset["scaler"] if "scaler" in dataset else None
            loader = LoaderWrapper(dataset, features, device=self.device, type = type)
            self.loaders.append(loader)
            self.scalers.update(loader.scalers)

    def set_val_loaders(self):
        if "validation" not in self.config:
            self.val_loaders = None
            return
        self.val_loaders = []
        features = self.config["features"]
        for dataset in self.config["validation"]:
            dataset = self.config["validation"][dataset]
            loader = LoaderWrapper(dataset, features, device=self.device, scalers=self.scalers)
            self.val_loaders.append(loader)

    def set_test_loaders(self):
        if "test" not in self.config:
            self.test_loaders = None
            return
        self.test_loaders = []
        features = self.config["features"]
        for dataset in self.config["test"]:
            dataset = self.config["test"][dataset]
            loader = LoaderWrapper(dataset, features, device=self.device, scalers=self.scalers)
            self.test_loaders.append(loader)
        
    def set_rule_to_data_loader_mapping(self):
        rule_to_loader_mapping = {}

        for rule in self.rules:
            variables = rule.variables()
            for v in variables:
                for loader in self.loaders:
                    if str(v) in loader.variables or str(v) in loader.targets:
                        if rule in rule_to_loader_mapping:
                            rule_to_loader_mapping[rule].append(loader)
                        else:
                            rule_to_loader_mapping[rule] = [loader]

        self.rule_to_data_loader_mapping = rule_to_loader_mapping
    
    def loss(self, rule_outputs):
        input = []
        for weight, rule_output in zip(self.rule_weights, rule_outputs):
            for _ in range(weight):
                input.append(rule_output)
                                
        sat_agg_value = sat_agg_op(
            *input,
        )
        loss = 1.0 - sat_agg_value

        assert torch.isfinite(loss).all(), f"Loss contains invalid values: {loss}"

        return loss
        
    def parameters(self):
        params = []
        for model in self.predicates.values():
            if hasattr(model, 'parameters'):
                params += list(model.parameters())
        return params
    
    def partition_data(self, var_mapping, batch, loader):
        for k, v in self.constant_mapping.items(): 
            var_mapping[k] = v.to(self.device)

        *batch, = batch 

        for i, var in enumerate(loader.variables):
            var_mapping[var] = batch[i].to(self.device)

        for i, target in enumerate(loader.targets):
            var_mapping[target] = batch[i + len(loader.variables)].to(self.device)
            
    def compute_validation_loss(self):
        with torch.no_grad():
            rule_outputs = []
            for rule in self.rules:
                rule_output = []
                var_mapping = {}
                if self.val_loaders:
                    for loader in self.val_loaders:
                        for batch in loader:
                            self.partition_data(var_mapping, batch, loader)
                            rule_output.append(rule(var_mapping))
                else:
                    rule_output.append(rule(var_mapping))
                rule_outputs.append(torch.mean(torch.stack(rule_output)))
            validation_loss = self.loss(rule_outputs)
        return validation_loss
    
    def compute_test_loss(self):
        with torch.no_grad():
            rule_outputs = []
            for rule in self.rules:
                rule_output = []
                var_mapping = {}
                if self.test_loaders:
                    for loader in self.test_loaders:
                        for batch in loader:
                            self.partition_data(var_mapping, batch, loader)
                            rule_output.append(rule(var_mapping))
                else:
                    rule_output.append(rule(var_mapping))
                rule_outputs.append(torch.mean(torch.stack(rule_output)))
            test_loss = self.loss(rule_outputs)
        return test_loss

    def optimize(self, num_epochs=10, log_steps=10, lr=0.001, early_stopping=False, patience=5, min_delta=0.0, weight_decay=0.0, verbose = True):
    
        try:
            self.optimizer = torch.optim.Adam(self.parameters(), lr=lr, weight_decay=weight_decay)
        except Exception as e:
            print(e)
            print("No parameters to optimize")
            return

        all_loaders = set(loader for loaders in self.rule_to_data_loader_mapping.values() if loaders is not None for loader in loaders if loader is not None)
        combined_loader = CombinedDataLoader([loader for loader in all_loaders if loader is not None])
        best_val_loss = float('inf')
        epochs_no_improve = 0

        for epoch in range(num_epochs):
            for _ in range(len(combined_loader)):
                rule_outputs = []
                current_batches = next(combined_loader)

                for rule in self.rules:
                    loaders = self.rule_to_data_loader_mapping[rule]
                    var_mapping = {}
                
                    if loaders == None:
                        rule_outputs.append(rule(var_mapping))
                        continue
                        
                    for loader in loaders:
                        batch = current_batches[loader]
                        self.partition_data(var_mapping, batch, loader)
                    
                    rule_output = rule(var_mapping)
                    rule_outputs.append(rule_output)
                            
                self.optimizer.zero_grad()
                loss = self.loss(rule_outputs)
                loss.backward()
                self.optimizer.step()

            if epoch % log_steps == 0 and verbose:
                validation_loss = self.compute_validation_loss() if self.val_loaders else None

                for rule, outcome in zip(self.rules, rule_outputs):
                    print(f"Rule: {rule}, Outcome: {outcome}")
                
                if validation_loss is not None:
                    print(f"Epoch {epoch + 1}/{num_epochs}, Train Loss: {loss.item()}, Validation Loss: {validation_loss.item()}")
                else:
                    print(f"Epoch {epoch + 1}/{num_epochs}, Train Loss: {loss.item()}")
                print()

            # Early Stopping Logic
            if early_stopping and self.val_loaders and verbose:
                validation_loss = self.compute_validation_loss()
                if validation_loss + min_delta < best_val_loss:
                    best_val_loss = validation_loss
                    epochs_no_improve = 0
                else:
                    epochs_no_improve += 1

                if epochs_no_improve >= patience:
                    for rule, outcome in zip(self.rules, rule_outputs):
                        print(f"Rule: {rule}, Outcome: {outcome}")
                    print(f"Early stopping at Epoch {epoch + 1}/{num_epochs}, Train Loss: {loss.item()}, Validation Loss: {validation_loss.item()}")
                    break



from ltn_imp.fuzzy_operators.connectives import AndConnective, OrConnective, NotConnective, ImpliesConnective, IffConnective, EqConnective, AddConnective, SubtractConnective, MultiplyConnective, DivideConnective, LessThanConnective, MoreThanConnective, LessThanOrEqualConnective, MoreThanOrEqualConnective, NegativeConnective
from ltn_imp.fuzzy_operators.predicates import Predicate
from ltn_imp.fuzzy_operators.quantifiers import ForallQuantifier, ExistsQuantifier
from ltn_imp.visitor import Visitor, make_visitable
import ltn_imp.fuzzy_operators.connectives as Connectives

import torch
from torchviz import make_dot
from collections import defaultdict
from ltn_imp.parsing.parser_generator import LTNParser
from ltn_imp.parsing.expressions import *
import os

make_visitable(Expression)    

def get_subclass_with_prefix(module, superclass: type, prefix: str = "default"):
    prefix = prefix.lower()
    for k in dir(module):
        obj = getattr(module, k)
        if isinstance(obj, type) and k.lower().startswith(prefix) and issubclass(obj, superclass):
            return obj()
    
    raise KeyError(f'No subtype of {superclass} found in module {module} with prefix "{prefix}"')

class ConvertedExpression:
    def __init__(self, expression, converted, visitor, device=None):
        self.expression = expression
        self.converted = converted
        self.visitor = visitor
        self.device = device

        if hasattr(self.expression, "left"):
            self.left = expression.left

        if hasattr(self.expression, "right"):
            self.right = expression.right

        if hasattr(self.expression, "term"):
            self.term = expression.term

        if hasattr(self.expression, "variable"):
            self.variable = expression.variable

        if hasattr(self.expression, "constant"):
            self.constant = expression.constant

        if hasattr(self.expression, "function"):
            self.function = expression.function

        if hasattr(self.expression, "quantifier"):
            self.quantifier = expression.quantifier

        if hasattr(self.expression, "feature"):
            self.feature = expression.feature

        if hasattr(self.expression, "args"):
            self.args = expression.args

        if hasattr(self.expression, "variables"):
            self.variables = expression.variables


    def __call__(self, *args, **kwargs):
        try:
            result = self.converted(*args, **kwargs)
            if isinstance(result, torch.Tensor):
                result = result.to(self.device)
            return result
        except Exception as e:
            print(f"For Expression {self.expression} this error occurred: {e}")
            raise e
        
    def __str__(self):
        return str(self.expression)
    
    def __repr__(self):
        return f"{self.expression}"

    def comp_graph(self, var_mapping):
        # Get the final result
        result = self(var_mapping)
        # Collect all intermediate results with requires_grad info
        params = {
            f'{name}_{i} (requires_grad={param.requires_grad})': param 
            for name, param_list in self.visitor.intermediate_results.items() 
            for i, param in enumerate(param_list)
        }
        # Visualize the computation graph
        dot = make_dot(result, params=params)
        return dot
    
class ExpressionVisitor(Visitor):
    def __init__(self, yaml, predicates, functions, connective_impls=None, quantifier_impls=None, declarations=None, declarers=None, device=torch.device("cpu"), scalers = None):
        connective_impls = connective_impls or {}
        quantifier_impls = quantifier_impls or {}

        self.predicates = predicates
        self.functions = functions

        self.declarations = declarations if declarations is not None else {}
        self.declarers = declarers if declarers is not None else {}
        self.device = device
        self.scalers = scalers

        self.yaml = yaml

        And = get_subclass_with_prefix(module=Connectives, superclass=AndConnective, prefix=connective_impls.get('and', 'default'))
        Or = get_subclass_with_prefix(module=Connectives, superclass=OrConnective, prefix=connective_impls.get('or', 'default'))
        Not = get_subclass_with_prefix(module=Connectives, superclass=NotConnective, prefix=connective_impls.get('not', 'default'))
        Implies = get_subclass_with_prefix(module=Connectives, superclass=ImpliesConnective, prefix=connective_impls.get('implies', 'default'))
        Equiv = get_subclass_with_prefix(module=Connectives, superclass=IffConnective, prefix=connective_impls.get('iff', 'default'))

        Eq_Regression = get_subclass_with_prefix(module=Connectives, superclass=EqConnective, prefix=connective_impls.get('eq_reg', 'sqrt'))
        Eq_Classification = get_subclass_with_prefix(module=Connectives, superclass=EqConnective, prefix=connective_impls.get('eq_class', 'tan'))        

        Exists = ExistsQuantifier(method=quantifier_impls.get('exists', 'pmean'))
        Forall = ForallQuantifier(method=quantifier_impls.get('forall', 'pmean_error'))

        Add = get_subclass_with_prefix(module=Connectives, superclass=AddConnective, prefix=functions.get('add', 'default'))
        Subtract = get_subclass_with_prefix(module=Connectives, superclass=SubtractConnective, prefix=functions.get('sub', 'default'))
        Multiply = get_subclass_with_prefix(module=Connectives, superclass=MultiplyConnective, prefix=functions.get('mul', 'default'))
        Divide = get_subclass_with_prefix(module=Connectives, superclass=DivideConnective, prefix=functions.get('div', 'default'))

        LessThan = get_subclass_with_prefix(module=Connectives, superclass=LessThanConnective, prefix=predicates.get('lt', 'default'))
        MoreThan = get_subclass_with_prefix(module=Connectives, superclass=MoreThanConnective, prefix=predicates.get('gt', 'default'))
        LessThanEqual = get_subclass_with_prefix(module=Connectives, superclass=LessThanOrEqualConnective, prefix=predicates.get('le', 'default'))
        MoreThanEqual = get_subclass_with_prefix(module=Connectives, superclass=MoreThanOrEqualConnective, prefix=predicates.get('ge', 'default'))

        Negative = get_subclass_with_prefix(module=Connectives, superclass=NegativeConnective, prefix=connective_impls.get('neg', 'default'))
        
        self.connective_map = {
            AndExpression: And,
            OrExpression: Or,
            ImpExpression: Implies,
            IffExpression: Equiv,
            NegatedExpression: Not,
            NegativeExpression: Negative,
            EqualityExpression: Eq_Regression,
            DirectEqualityExpression : Eq_Classification, 
            AdditionExpression: Add,
            SubtractionExpression: Subtract,
            MultiplicationExpression: Multiply,
            DivisionExpression: Divide,
            LessThanExpression: LessThan,
            MoreThanExpression: MoreThan,
            LessEqualExpression: LessThanEqual, # Still Problematic
            MoreEqualExpression: MoreThanEqual  # Still Problematic
        }

        self.quantifier_map = {
            ExistsExpression: Exists,
            ForallExpression: Forall
        }

        self.intermediate_results = defaultdict(list)

    def handle_predicate(self, variables, functor, var_mapping, expression):
        inputs = []
        predicate = Predicate(self.predicates[functor])

        to_be_declared = None

        for i, var in enumerate(variables):
            if type(var) != VariableExpression and type(var) != ConstantExpression:
                var = self.visit(var)(var_mapping)
                inputs.append(var)
                continue
    
            var = str(var)

            if var in var_mapping:
                inputs.append(var_mapping[var])
                
            elif var in self.declarations:
                if self.declarers[var] == str(expression):
                    to_be_declared = variables[i:]
                    break
                else:
                    value = self.declarations[var]
                    inputs.append(value)
            else:
                to_be_declared = variables[i:]
                break

        for input in inputs:
           input = input.to(self.device)

        results = predicate(*inputs)
        
        if not isinstance(results, tuple) and results.dim() == 0:
            results = results.unsqueeze(0)
            
        if to_be_declared is not None:
            for i, var in enumerate(to_be_declared):
                value = results[i] if len(to_be_declared) > 1 else results
                self.declarations[str(var)] = value
                self.declarers[str(var)] = str(expression)
                
            self.intermediate_results[functor].append(torch.tensor([1.0], requires_grad=True, device=self.device))
            return torch.tensor([1.0], requires_grad=True, device=self.device)

        # Ensuring results is a tensor
        if isinstance(results, (list, tuple)):
            results_tensor = torch.stack([res if isinstance(res, torch.Tensor) else torch.tensor(res, device=self.device) for res in results])
        else:
            results_tensor = results

        self.intermediate_results[functor].append(results_tensor)
        return results_tensor

    def visit_ApplicationExpression(self, expression):
        variables = [arg for arg in expression.args]
        functor = expression
        
        while hasattr(functor, 'function'):
            functor = functor.function

        functor = str(functor)
        
        if functor in self.predicates:
            return ConvertedExpression(expression, lambda var_mapping: self.handle_predicate(variables, functor, var_mapping, expression), self)
        else:
            raise ValueError(f"Unknown functor: {functor}")

    def delay_execution(self, left, right, var_mapping, connective):
        try: # Right side might be declaring a variable in the left 
            left_value = left(var_mapping)
            right_value = right(var_mapping)
        except:
            right_value = right(var_mapping)
            left_value = left(var_mapping)
        return connective(left_value, right_value)
        
    def visit_BinaryExpression(self, expression):
        connective = self.connective_map.get(type(expression))
        if connective:
            left = self.visit(expression.left)
            right = self.visit(expression.right)
            return ConvertedExpression(expression, lambda var_mapping: self.delay_execution(left, right, var_mapping, connective), self)
        else:
            raise NotImplementedError(f"Unsupported binary expression type: {type(expression)}")

    def visit_NegatedExpression(self, expression):
        connective = self.connective_map.get(type(expression))
        term = self.visit(expression.term)
        return ConvertedExpression(expression, lambda var_mapping: connective(term.converted(var_mapping)), self)
    
    def visit_NegativeExpression(self, expression):
        connective = self.connective_map.get(type(expression))
        term = self.visit(expression.term)
        return ConvertedExpression(expression, lambda var_mapping: connective(term.converted(var_mapping)), self)

    def visit_QuantifiedExpression(self, expression):
        quantifier = self.quantifier_map.get(type(expression))
        if quantifier:
            term = self.visit(expression.term)
            return ConvertedExpression(expression, lambda variable_mapping: quantifier(term.converted(variable_mapping)), self)
        else:
            raise NotImplementedError(f"Unsupported quantifier expression type: {type(expression)}")
                
    def handle_variable(self, variable_mapping, expression):
        var = expression.variable
        if str(var) in variable_mapping:
            return variable_mapping[str(var)]
        elif str(var) in self.declarations:
            return self.declarations[str(var)]
        else:
            raise KeyError(f"Variable {var} is not recognized")
    
    def visit_VariableExpression(self, expression):
        return ConvertedExpression(expression, lambda variable_mapping: self.handle_variable(variable_mapping, expression), self)

    def handle_constant(self, variable_mapping, expression):
        if str(expression) in variable_mapping:
            return variable_mapping[str(expression)]
        else:
            return torch.tensor(float(str(expression)),requires_grad=True, device=self.device)
                
    def visit_ConstantExpression(self, expression):
        return ConvertedExpression(expression, lambda variable_mapping: self.handle_constant(variable_mapping, expression), self)

    def reverse_scaling(self, scaler, tensor_data):    

        if hasattr(scaler, 'mean_') and hasattr(scaler, 'scale_'):
            # For StandardScaler
            scaler_mean = torch.tensor(scaler.mean_, dtype=torch.float32).to(tensor_data.device)
            scaler_scale = torch.tensor(scaler.scale_, dtype=torch.float32).to(tensor_data.device)            
            scaler_mean = scaler_mean[0]  # Extract mean for the feature column
            scaler_scale = scaler_scale[0]  # Extract scale for the feature column
            tensor_data_reversed = tensor_data * scaler_scale + scaler_mean
        
        elif hasattr(scaler, 'data_min_') and hasattr(scaler, 'data_max_'):
            # For MinMaxScaler
            scaler_min = torch.tensor(scaler.data_min_, dtype=torch.float32).to(tensor_data.device)
            scaler_max = torch.tensor(scaler.data_max_, dtype=torch.float32).to(tensor_data.device)            
            scaler_min = scaler_min[0]
            scaler_max = scaler_max[0]
            tensor_data_reversed = tensor_data * (scaler_max - scaler_min) + scaler_min
            
        else:
            raise ValueError("Unsupported scaler type. Only StandardScaler and MinMaxScaler are supported.")
    
        return tensor_data_reversed


    def handle_indexing(self, variable_mapping, expression):
        feature = expression.feature
        variable = expression.variable
        if self.scalers:
            scaler = self.scalers[str(variable)][str(feature)] if str(variable) in self.yaml["instances"] else None
        else:
            scaler = None

        if str(variable) in variable_mapping:
            index = self.yaml["features"][str(variable)].index(feature)
            return self.reverse_scaling( scaler, self.visit(variable)(variable_mapping)[:, index] )  if scaler else self.visit(variable)(variable_mapping)[:, index]
        else:
            if str(feature).isdigit():
                feature = int(feature)
                return self.reverse_scaling( scaler, self.visit(variable)(variable_mapping)[:, feature] ) if scaler else self.visit(variable)(variable_mapping)[:, feature]
            else:
                raise KeyError(f"Variable {variable} not recognized")

    def visit_IndexExpression(self, expression):
        return ConvertedExpression(expression, lambda variable_mapping: self.handle_indexing(variable_mapping, expression), self)

class LTNConverter:
    def __init__(self,yaml = None, scalers = {}, predicates={}, functions={}, connective_impls=None, quantifier_impls=None, declarations={}, declarers={}, device=torch.device("cpu")):
        self.predicates = predicates
        self.functions = functions
        self.connective_impls = connective_impls
        self.quantifier_impls = quantifier_impls
        self.declarations = declarations
        self.declarers = declarers
        self.expression = None
        parser_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "fol_parser.ebnf")
        self.parser = LTNParser(parser_path)
        self.yaml = yaml
        self.device = device
        self.scalers = scalers

        self.visitor = ExpressionVisitor(
            self.yaml,
            self.predicates, 
            self.functions, 
            connective_impls=self.connective_impls, 
            quantifier_impls=self.quantifier_impls, 
            declarations=self.declarations, 
            declarers=self.declarers,
            device=self.device,
            scalers = self.scalers,
        )
        
    def parse(self, expression):
        expression = self.parser.parse(expression)
        self.expression = expression
        return expression
        
    def __call__(self, expression):
        expression = self.parser.parse(expression)
        self.expression = expression
        return ConvertedExpression(self.expression, expression.accept(self.visitor), self.visitor, device=self.device)



from itertools import cycle
from torch.utils.data import Dataset, DataLoader
import pandas as pd
import torch
from sklearn.preprocessing import StandardScaler, MinMaxScaler

class DynamicDataset(Dataset):
    def __init__(self, config, features, device=None, scalers=None, type = None):
        self.config = config
        self.features = features
        self.device = device  # Store the device
        self.data = pd.read_csv(config["path"], index_col=0)
        self.batch_size = config["batch_size"]
        
        self.instance_features = {instance: [str(feat) for feat in features[instance]] for instance in config["instances"]}
        self.target_features = {target: [str(feat) for feat in features[target]] for target in config["targets"]}

        if any(self.data.index.name in self.instance_features[instance] for instance in self.config["instances"]):
            self.data = pd.read_csv(config["path"])

        if type == "standard":
            scaler = StandardScaler

        if type == "minmax":
            scaler = MinMaxScaler

        if scalers is None and type is not None:
            self.scalers = {}
            for instance in self.config["instances"]:
                self.scalers[instance] = {}
                for feature in self.instance_features[instance]:
                    self.scalers[instance][feature] = scaler()
                    self.data[feature] = self.scalers[instance][feature].fit_transform(self.data[feature].values.reshape(-1, 1))

        elif scalers:
            self.scalers = scalers
            for instance in self.config["instances"]:
                for feature in self.instance_features[instance]:
                    self.data[feature] = self.scalers[instance][feature].fit_transform(self.data[feature].values.reshape(-1, 1))

        elif type is None:
            self.scalers = {}
            return 


    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, idx):
        batch = []

        # Get instance data
        for instance in self.config["instances"]:
            instance_data = self.data[self.instance_features[instance]].iloc[idx]
            batch.append(instance_data.values)

        # Get target data
        for target in self.config["targets"]:
            target_data = self.data[self.target_features[target]].iloc[idx]
            batch.append(target_data.values)

        # Convert to tensors and move to the specified device
        batch = [torch.tensor(item, dtype=torch.float32).to(self.device) for item in batch]
        
        return tuple(batch)
class LoaderWrapper:
    def __init__(self, config, features, device=None, scalers = None, type = None):
        self.variables = config["instances"]
        self.targets = config["targets"]
        self.loader = DataLoader(DynamicDataset(config, features, device=device, scalers=scalers, type = type), batch_size=config["batch_size"], shuffle=True)
        self.scalers = self.loader.dataset.scalers

    def __iter__(self):
        self.iter_loader = iter(self.loader)
        return self
    
    def __next__(self):
        return next(self.iter_loader)
    
    def __len__(self):
        return len(self.loader)
    
    def __repr__(self) -> str:
        return f"<Loader>:({self.variables} -> {self.targets})"
    
class CombinedDataLoader:
    def __init__(self, loaders):
        self.loaders = loaders
        self.iters = {loader: cycle(loader) for loader in loaders}
        self.current_batches = {loader: None for loader in loaders}
        self.step()

        if loaders != []:
            self.max_length = max(len(loader) for loader in loaders)
        else:
            self.max_length = 1

    def __iter__(self):
        return self

    def __len__(self):
        return self.max_length
    
    def __next__(self):
        return self.current_batches

    def step(self):
        for loader in self.loaders:
            self.current_batches[loader] = next(self.iters[loader])


